// SPDX-FileCopyrightText: 2009 The Go Authors
// SPDX-FileCopyrightText: 2022-2023 nervuri <https://nervuri.net/contact>
//
// SPDX-License-Identifier: BSD-3-Clause

// Based on Go's crypto/tls library:
// https://github.com/golang/go/blob/master/src/crypto/tls/handshake_messages.go#L69
// https://cs.opensource.google/go/go/+/refs/tags/go1.18:src/crypto/tls/handshake_messages.go;l=69

package clienthello

import (
	"crypto/md5"
	"encoding/binary"
	"encoding/hex"
	"golang.org/x/crypto/cryptobyte"
	"log"
	"strconv"
	"strings"
)

// Check if this is a GREASE value (RFC 8701).
func isGREASE(value uint16) bool {
	// Values for: cipher suites, ALPN,
	// extensions, named groups, signature algorithms and TLS versions:
	greaseValues := [16]uint16{
		0x0a0a, 0x1a1a, 0x2a2a, 0x3a3a, 0x4a4a, 0x5a5a, 0x6a6a, 0x7a7a,
		0x8a8a, 0x9a9a, 0xaaaa, 0xbaba, 0xcaca, 0xdada, 0xeaea, 0xfafa,
	}
	// Run the check
	for _, greaseValue := range greaseValues {
		if value == greaseValue {
			return true
		}
	}
	return false
}

// readUint8LengthPrefixed acts like s.ReadUint8LengthPrefixed, but targets a
// []byte instead of a cryptobyte.String.
func readUint8LengthPrefixed(s *cryptobyte.String, out *[]byte) bool {
	return s.ReadUint8LengthPrefixed((*cryptobyte.String)(out))
}

// readUint16LengthPrefixed acts like s.ReadUint16LengthPrefixed, but targets a
// []byte instead of a cryptobyte.String.
func readUint16LengthPrefixed(s *cryptobyte.String, out *[]byte) bool {
	return s.ReadUint16LengthPrefixed((*cryptobyte.String)(out))
}

type byteSlice []byte

func (bs byteSlice) MarshalText() ([]byte, error) {
	// Used in json.Marshal() to convert []byte to hex string.
	return []byte(hex.EncodeToString(bs)), nil
}

type highlights struct {
	//SupportedTLSVersions     []uint16
	GMTUnixTime                uint32 `json:"gmt_unix_time"` // first 4 bytes of client random
	SecureRenegotiationSupport bool   `json:"secure_renegotiation_support"`
	OCSPStaplingSupport        bool   `json:"ocsp_stapling_support"`
	SCTSupport                 bool   `json:"sct_support"`
	// Go's crypto/tls server does not support early data.
	EarlyData bool      `json:"-"` // don't include in JSON
	JA3       string    `json:"ja3"`
	JA3MD5    byteSlice `json:"ja3_md5"`
}

type ClientHelloMsg struct {
	Raw                    byteSlice           `json:"raw"`
	RecordHeaderTLSVersion TLSVersion          `json:"record_header_tls_version"` // TLSv1.0 (769)
	TLSVersion             TLSVersion          `json:"client_tls_version"`        // TLSv1.2 (771)
	Random                 byteSlice           `json:"random"`
	SessionID              byteSlice           `json:"session_id"`
	CipherSuites           []CipherSuite       `json:"cipher_suites"`
	CompressionMethods     []CompressionMethod `json:"compression_methods"`
	Extensions             []Extension         `json:"extensions"`
	Highlights             highlights          `json:"highlights"`
}

func (m *ClientHelloMsg) Unmarshal(data []byte) bool {
	*m = ClientHelloMsg{Raw: data}
	s := cryptobyte.String(data)

	var recordHeaderTLSVersion uint16
	var tlsVersion uint16
	var random []byte
	var sessionID []byte
	if !s.Skip(1) || !s.ReadUint16(&recordHeaderTLSVersion) || !s.Skip(2) ||
		!s.Skip(4) || // message type and uint24 length field
		!s.ReadUint16(&tlsVersion) || !s.ReadBytes(&random, 32) ||
		!readUint8LengthPrefixed(&s, &sessionID) {
		return false
	}
	m.RecordHeaderTLSVersion = recordHeaderTLSVersion
	m.TLSVersion = tlsVersion
	m.Random = random
	m.Highlights.GMTUnixTime = binary.BigEndian.Uint32(random[0:4])

	m.SessionID = sessionID

	m.CipherSuites = []CipherSuite{}
	var cipherSuites cryptobyte.String
	if !s.ReadUint16LengthPrefixed(&cipherSuites) {
		return false
	}
	m.Highlights.SecureRenegotiationSupport = false
	for !cipherSuites.Empty() {
		var suite uint16
		if !cipherSuites.ReadUint16(&suite) {
			return false
		}
		if suite == scsvRenegotiation {
			m.Highlights.SecureRenegotiationSupport = true
		}
		m.CipherSuites = append(m.CipherSuites, suite)
	}

	var compressionMethods []uint8
	if !readUint8LengthPrefixed(&s, &compressionMethods) {
		return false
	}
	for _, cm := range compressionMethods {
		m.CompressionMethods = append(m.CompressionMethods, cm)
	}

	if s.Empty() {
		// ClientHello is optionally followed by extension data
		return true
	}

	var extensions cryptobyte.String
	if !s.ReadUint16LengthPrefixed(&extensions) || !s.Empty() {
		return false
	}

	for !extensions.Empty() {
		var extCode uint16
		var extData cryptobyte.String
		if !extensions.ReadUint16(&extCode) ||
			!extensions.ReadUint16LengthPrefixed(&extData) {
			return false
		}
		var extension = GetExtensionInfo(extCode)
		extension.Data.Raw = []byte(extData)

		switch extension.Code {
		case extensionServerName:
			// Server Name Indication
			// RFC 6066, Section 3
			var nameList cryptobyte.String
			if !extData.ReadUint16LengthPrefixed(&nameList) || nameList.Empty() {
				return false
			}
			for !nameList.Empty() {
				var nameType uint8
				var serverName cryptobyte.String
				if !nameList.ReadUint8(&nameType) ||
					!nameList.ReadUint16LengthPrefixed(&serverName) ||
					serverName.Empty() {
					return false
				}
				if nameType != 0 {
					continue
				}
				if len(extension.Data.ServerName) != 0 {
					// Multiple names of the same name_type are prohibited.
					return false
				}
				sn := string(serverName)
				extension.Data.ServerName = sn
				// An SNI value may not include a trailing dot.
				if strings.HasSuffix(sn, ".") {
					return false
				}
			}
		case extensionNextProtoNeg:
			// Next Protocol Negotiation
			// draft-agl-tls-nextprotoneg-04 (not IANA assiged)
			extension.Name = "next_protocol_negotiation"
		case extensionStatusRequest:
			// Certificate Status Request
			// RFC 4366, Section 3.6
			var ignored cryptobyte.String
			if !extData.ReadUint8(&extension.Data.StatusType) ||
				!extData.ReadUint16LengthPrefixed(&ignored) || // responder_id_list
				!extData.ReadUint16LengthPrefixed(&ignored) { // request_extensions
				return false
			}
			m.Highlights.OCSPStaplingSupport = extension.Data.StatusType == statusTypeOCSP
		case extensionSupportedGroups:
			// Supported Groups (former name: elliptic_curves)
			// RFC 4492, sections 5.1.1 and RFC 8446, Section 4.2.7
			var groups cryptobyte.String
			if !extData.ReadUint16LengthPrefixed(&groups) || groups.Empty() {
				return false
			}
			extension.Data.SupportedGroups = make([]NamedGroup, 0)
			for !groups.Empty() {
				var group uint16
				if !groups.ReadUint16(&group) {
					return false
				}
				extension.Data.SupportedGroups = append(
					extension.Data.SupportedGroups, group)
			}
		case extensionSupportedPointFormats:
			// Supported Point Formats
			// RFC 4492, Section 5.1.2
			var supportedPointFormats []uint8
			if !readUint8LengthPrefixed(&extData, &supportedPointFormats) ||
				len(supportedPointFormats) == 0 {
				return false
			}
			extension.Data.SupportedPointFormats = make(
				[]ECPointFormat, len(supportedPointFormats))
			for i, pointFormatCode := range supportedPointFormats {
				extension.Data.SupportedPointFormats[i] = pointFormatCode
			}
		case extensionSessionTicket:
			// Session Ticket
			// RFC 5077, Section 3.2
		case extensionSignatureAlgorithms:
			// Signature Algorithms
			// RFC 5246, Section 7.4.1.4.1
			extension.Data.SupportedSignatureAlgorithms = []SignatureScheme{}
			var sigAndAlgs cryptobyte.String
			if !extData.ReadUint16LengthPrefixed(&sigAndAlgs) || sigAndAlgs.Empty() {
				return false
			}
			for !sigAndAlgs.Empty() {
				var sigAndAlg uint16
				if !sigAndAlgs.ReadUint16(&sigAndAlg) {
					return false
				}
				extension.Data.SupportedSignatureAlgorithms = append(
					extension.Data.SupportedSignatureAlgorithms, sigAndAlg)
			}
		case extensionSignatureAlgorithmsCert:
			// Signature Algorithms Cert
			// RFC 8446, Section 4.2.3
			extension.Data.SupportedSignatureAlgorithms = []SignatureScheme{}
			var sigAndAlgs cryptobyte.String
			if !extData.ReadUint16LengthPrefixed(&sigAndAlgs) || sigAndAlgs.Empty() {
				return false
			}
			for !sigAndAlgs.Empty() {
				var sigAndAlg uint16
				if !sigAndAlgs.ReadUint16(&sigAndAlg) {
					return false
				}
				extension.Data.SupportedSignatureAlgorithms = append(
					extension.Data.SupportedSignatureAlgorithms, sigAndAlg)
			}
		case extensionRenegotiationInfo:
			// Renegotiation Indication
			// RFC 5746, Section 3.2
			if !readUint8LengthPrefixed(
				&extData,
				&extension.Data.RenegotiationInfo) {
				return false
			}
			m.Highlights.SecureRenegotiationSupport = true
		case extensionALPN:
			// Application-Layer Protocol Negotiation
			// RFC 7301, Section 3.1
			var protoList cryptobyte.String
			if !extData.ReadUint16LengthPrefixed(&protoList) || protoList.Empty() {
				return false
			}
			for !protoList.Empty() {
				var proto cryptobyte.String
				if !protoList.ReadUint8LengthPrefixed(&proto) || proto.Empty() {
					return false
				}
				extension.Data.ALPNProtocols = append(
					extension.Data.ALPNProtocols, string(proto))
			}
		case extensionSCT:
			// Signed Certificate Timestamp
			// RFC 6962, Section 3.3.1
			m.Highlights.SCTSupport = true
		case extensionSupportedVersions:
			// Supported Versions
			// RFC 8446, Section 4.2.1
			var versList cryptobyte.String
			extension.Data.SupportedVersions = []TLSVersion{}
			if !extData.ReadUint8LengthPrefixed(&versList) || versList.Empty() {
				return false
			}
			for !versList.Empty() {
				var vers uint16
				if !versList.ReadUint16(&vers) {
					return false
				}
				extension.Data.SupportedVersions = append(
					extension.Data.SupportedVersions, vers)
			}
		case extensionCookie:
			// Cookie
			// RFC 8446, Section 4.2.2
			var cookie []byte
			if !readUint16LengthPrefixed(&extData, &cookie) ||
				len(extension.Data.Cookie) == 0 {
				return false
			}
			extension.Data.Cookie = cookie
		case extensionKeyShare:
			// Key Share
			// RFC 8446, Section 4.2.8
			var clientShares cryptobyte.String
			if !extData.ReadUint16LengthPrefixed(&clientShares) {
				return false
			}
			for !clientShares.Empty() {
				var ks KeyShare
				var ksGroup uint16
				var ksData []byte
				if !clientShares.ReadUint16(&ksGroup) ||
					!readUint16LengthPrefixed(&clientShares, &ksData) ||
					len(ksData) == 0 {
					return false
				}
				ks.Group = ksGroup
				ks.Data = ksData
				extension.Data.KeyShares = append(extension.Data.KeyShares, ks)
			}
		case extensionEarlyData:
			// Early Data Indication
			// RFC 8446, Section 4.2.10
			m.Highlights.EarlyData = true
		case extensionPSKModes:
			// Pre-Shared Key Exchange Modes
			// RFC 8446, Section 4.2.9
			var pskModes []uint8
			if !readUint8LengthPrefixed(&extData, &pskModes) {
				return false
			}
			for _, pskMode := range pskModes {
				extension.Data.PSKModes = append(extension.Data.PSKModes, pskMode)
			}
		case extensionPreSharedKey:
			// Pre-Shared Key
			// RFC 8446, Section 4.2.11
			if !extensions.Empty() {
				return false // pre_shared_key must be the last extension
			}
			var identities cryptobyte.String
			if !extData.ReadUint16LengthPrefixed(&identities) || identities.Empty() {
				return false
			}
			for !identities.Empty() {
				var psk PSKIdentity
				var identity []byte
				if !readUint16LengthPrefixed(&identities, &identity) ||
					!identities.ReadUint32(&psk.ObfuscatedTicketAge) ||
					len(identity) == 0 {
					return false
				}
				psk.Identity = identity
				extension.Data.PSKIdentities = append(extension.Data.PSKIdentities, psk)
			}
			var binders cryptobyte.String
			if !extData.ReadUint16LengthPrefixed(&binders) || binders.Empty() {
				return false
			}
			for !binders.Empty() {
				var binder []byte
				if !readUint8LengthPrefixed(&binders, &binder) ||
					len(binder) == 0 {
					return false
				}
				extension.Data.PSKBinders = append(extension.Data.PSKBinders, binder)
			}
		case extensionPadding:
			// Padding
			// RFC 7685
			extension.Data.Length = uint16(len(extData))
		case extensionCompressCertificate:
			// Certificate Compression
			// RFC 8879
			// TODO: parse - https://www.rfc-editor.org/rfc/rfc8879.html
		case extensionApplicationSettings:
			// Application-Layer Protocol Settings
			// draft-vvv-tls-alps (not IANA assiged)
			extension.Name = " application_settings"
			// TODO: parse - https://datatracker.ietf.org/doc/html/draft-vvv-tls-alps
		}

		m.Extensions = append(m.Extensions, extension)

		if extension.Name == "" {
			log.Println("Unknown extension:", extension.Code)
		}

		//if !extData.Empty() &&
		//	extension.Name != "GREASE" &&
		//	extension.Code != extensionPadding {
		//	log.Printf("Extension %d data not read.\n", extension.Code)
		//}
	}

	// JA3
	var ja3 strings.Builder
	var supportedGroups []NamedGroup
	var supportedPointFormats []ECPointFormat
	ja3.WriteString(strconv.FormatUint(uint64(m.TLSVersion.(uint16)), 10) + ",")
	for i, cs := range m.CipherSuites {
		if !isGREASE(cs.(uint16)) { // ignore GREASE values
			ja3.WriteString(strconv.FormatUint(uint64(cs.(uint16)), 10))
			if i+1 != len(m.CipherSuites) { // if not last element, add "-"
				ja3.WriteString("-")
			}
		}
	}
	ja3.WriteString(",")
	for i, e := range m.Extensions {
		if !isGREASE(e.Code) { // ignore GREASE values
			ja3.WriteString(strconv.FormatUint(uint64(e.Code), 10))
			if i+1 != len(m.Extensions) { // if not last element, add "-"
				ja3.WriteString("-")
			}
			if e.Code == extensionSupportedGroups {
				supportedGroups = e.Data.SupportedGroups
			} else if e.Code == extensionSupportedPointFormats {
				supportedPointFormats = e.Data.SupportedPointFormats
			}
		}
	}
	ja3.WriteString(",")
	for i, g := range supportedGroups {
		if !isGREASE(g.(uint16)) { // ignore GREASE values
			ja3.WriteString(strconv.FormatUint(uint64(g.(uint16)), 10))
			if i+1 != len(supportedGroups) { // if not last element, add "-"
				ja3.WriteString("-")
			}
		}
	}
	ja3.WriteString(",")
	for i, pf := range supportedPointFormats {
		ja3.WriteString(strconv.FormatUint(uint64(pf.(uint8)), 10))
		if i+1 != len(supportedPointFormats) { // if not last element, add "-"
			ja3.WriteString("-")
		}
	}
	m.Highlights.JA3 = ja3.String()
	hash := md5.Sum([]byte(m.Highlights.JA3))
	m.Highlights.JA3MD5 = hash[:]

	return true
}

// Add information about TLS versions, cipher suites, etc.
func (m *ClientHelloMsg) AddInfo() {
	m.AddTLSVersionInfo()
	m.AddCipherSuiteInfo()
	m.AddSignatureSchemeInfo()
	m.AddNamedGroupInfo() // supported_groups
	m.AddECPointFormatInfo()
	m.AddCompressionMethodInfo()
	m.AddPSKModeInfo()
}
