// SPDX-FileCopyrightText: 2022-2023 nervuri <https://nervuri.net/contact>
//
// SPDX-License-Identifier: BSD-3-Clause

package clienthello

import (
	_ "embed"
	"encoding/csv"
	"fmt"
	"io"
	"log"
	"strconv"
	"strings"
)

// RFC 8446, Section 4.2.3
// https://www.rfc-editor.org/rfc/rfc8446.html#section-4.2.3
// https://www.iana.org/assignments/tls-parameters/tls-signaturescheme.csv
//
//go:embed signature-schemes.csv
var signatureSchemesCSV string

var SignatureSchemeList = parseSignatureSchemesCSV()

type SignatureScheme any // (uint16 | SignatureSchemeInfo)

type SignatureSchemeInfo struct {
	Code        uint16 `json:"code"`
	HexCode     string `json:"hex_code"`
	Name        string `json:"name"`
	Recommended bool   `json:"-"`
	Reference   string `json:"-"`
}

func parseSignatureSchemesCSV() map[uint16]SignatureSchemeInfo {
	signatureSchemes := map[uint16]SignatureSchemeInfo{}
	r := csv.NewReader(strings.NewReader(signatureSchemesCSV))

	// Skip CSV header.
	_, err := r.Read()
	if err != nil {
		log.Fatal(err)
	}

	for {
		row, err := r.Read()
		if err == io.EOF {
			break
		} else if err != nil {
			log.Fatal(err)
		}
		var (
			val  = row[0]
			desc = row[1]
			rec  = row[2]
			ref  = row[3]
		)
		if desc == "Unassigned" {
			continue
		}
		if strings.HasPrefix(desc, "Reserved") && ref != "[RFC8701]" {
			continue
		}
		val = strings.Replace(val, "0x", "", 1)
		code, err := strconv.ParseUint(val, 16, 16)
		if err != nil {
			log.Println(err)
		}
		sigScheme := SignatureSchemeInfo{
			Code:      uint16(code),
			HexCode:   fmt.Sprintf("%04X", code),
			Name:      desc,
			Reference: ref,
		}
		if rec == "Y" {
			sigScheme.Recommended = true
		}
		if ref == "[RFC8701]" || isGREASE(sigScheme.Code) {
			// (We call isGREASE() because the IANA signature scheme registry
			// doesn't currently include GREASE values).
			sigScheme.Name = "GREASE"
		}

		signatureSchemes[sigScheme.Code] = sigScheme
	}
	return signatureSchemes
}

func GetSignatureSchemeInfo(sigSchemeCode uint16) SignatureSchemeInfo {
	info, found := SignatureSchemeList[sigSchemeCode]
	if !found {
		info = SignatureSchemeInfo{
			Code:    sigSchemeCode,
			HexCode: fmt.Sprintf("%04X", sigSchemeCode),
		}
	}
	return info
}

func (m *ClientHelloMsg) AddSignatureSchemeInfo() {
	for i, ext := range m.Extensions {
		if ext.Code == extensionSignatureAlgorithms ||
			ext.Code == extensionSignatureAlgorithmsCert {
			for j, sigAlg := range ext.Data.SupportedSignatureAlgorithms {
				m.Extensions[i].Data.SupportedSignatureAlgorithms[j] = GetSignatureSchemeInfo(
					sigAlg.(uint16))
			}
		}
	}
}
